from graphene import ObjectType, String, Field, JSONString, ResolveInfo, Int, NonNull, List
from ..types.element import GrapheneVertexType, GrapheneEdgeType, AnyField


class GremlinVertexMutationSchema:
    create_vertex = Field(GrapheneVertexType, label=String(required=True), namespace=String(),
                          properties=JSONString(required=True))
    update_vertex_by_id = Field(GrapheneVertexType, id=AnyField(required=True), properties=JSONString(required=True))
    remove_vertex_by_id = String(id=AnyField(required=True))
    remove_vertices = String(label=String(), namespace=String(), query=JSONString())

    def resolve_create_vertex(self, info: ResolveInfo, label: str, namespace: str, properties: str):
        data = info.context['request'].app.state.gremlin_client.vertex.create(
            label=label, namespace=namespace, properties=properties)
        return data.__dict__() if data else None

    def resolve_update_vertex_by_id(self, info: ResolveInfo, id: str, properties: str):
        data = info.context['request'].app.state.gremlin_client.vertex.update(id, properties=properties)
        return data.__dict__() if data else None

    def resolve_remove_vertex_by_id(self, info: ResolveInfo, id: str):
        data = info.context['request'].app.state.gremlin_client.vertex.delete_one(id)
        return data.__dict__() if data else None

    def resolve_remove_vertices(self, info: ResolveInfo, label: str = None, namespace: str = None, query: str = None):
        info.context['request'].app.state.gremlin_client.vertex.delete_many(label=label,
                                                                            namespace=namespace, query=query)
        return None



class GremlinEdgeMutationSchema:
    create_edge = Field(GrapheneEdgeType,
                        inv=String(required=True),
                        outv=String(required=True),
                        label=String(required=True),
                        name=String(),
                        properties=JSONString())
    update_edge_by_id = Field(GrapheneEdgeType, id=AnyField(required=True), properties=JSONString(required=True))
    remove_edge_by_id = String(id=AnyField(required=True))
    remove_edges = String(label=String(), namespace=String(), query=JSONString())

    def resolve_create_edge(self, info: ResolveInfo, label: str, namespace: str, properties: str, inv: str, outv: str):
        return info.context['request'].app.state.gremlin_client.edge.create(
            outv=outv, inv=inv, label=label, namespace=namespace, properties=properties
        )

    def resolve_update_edge_by_id(self, info: ResolveInfo, id: str, properties: str):
        data = info.context['request'].app.state.gremlin_client.edge.update(id, properties=properties)
        return data.__dict__() if data else None

    def resolve_remove_edge_by_id(self, info: ResolveInfo, id: str):
        data = info.context['request'].app.state.gremlin_client.edge.delete_one(id)
        return data.__dict__() if data else None

    def resolve_remove_edges(self, info: ResolveInfo, label: str = None, namespace: str = None, query: str = None):
        info.context['request'].app.state.gremlin_client.edge.delete_many(label=label,
                                                                          namespace=namespace, query=query)
        return None



class GremlinMutation(ObjectType, GremlinEdgeMutationSchema, GremlinVertexMutationSchema):
    pass



"""

"""

CREATE_VERTICES_SAMPLES = {
    "plant": {
        "label": "Planet",
        "properties": {
            "name": "Earth",
            "mass_in_kgs": float(5972000000000000000000000),
            "radius_in_kms": 6371
        }
    },
    "satellite": {
        "label": "Satellite",
        "properties": {
            "name": "Moon",
            "mass_in_kgs": int(73476730900000000000000)
        }
    }
}



import pytest
from .sample_payloads.client import CLIENT_RAW_QUERIES
import os

TEST_GRAPH_HOST = os.environ.get("TESTING_HOST")


class TestGremlinClientClient:

    @pytest.fixture
    def graph_client(self):
        from invana_engine.gremlin import GremlinClient
        return GremlinClient(f"{TEST_GRAPH_HOST}/gremlin")

    def test_raw_query(self, graph_client):
        for query_id, raw_query in CLIENT_RAW_QUERIES.items():
            response = graph_client.execute_query(raw_query)
            assert type(response) is list

    def test_drop_everything(self, graph_client):
        response = graph_client.delete_everything()
        # assert response is None



"""

"""


class InvalidVertexException(BaseException):
    pass


class InvalidPayloadException(BaseException):
    pass


class InvalidConnection(BaseException):
    pass


class InvalidQueryArguments(BaseException):
    pass


class EdgeDoesntExist(BaseException):
    pass


class VertexDoesntExist(BaseException):
    pass



import os

gremlin_server_url = os.environ.get("GREMLIN_SERVER_URL")
gremlin_server_username = os.environ.get("GREMLIN_SERVER_USERNAME")
gremlin_server_password = os.environ.get("GREMLIN_SERVER_PASSWORD")
shall_debug = os.environ.get("DEBUG", False)
server_port = os.environ.get("SERVER_PORT", 8200)



from .base import CRUDOperationsBase
import logging
from .core.element import VertexElement, EdgeElement
from ..server.utils import get_unique_items

logger = logging.getLogger(__name__)


class Vertex(CRUDOperationsBase):

    def create(self, label=None, namespace=None, properties=None, **kwargs):
        """

        :param label:
        :param namespace:
        :param properties:
        :param kwargs: not used
        :return:
        """
        logger.debug("Creating vertex with label {label} namespace {namespace}"
                     "and properties {properties}".format(label=label, namespace=namespace, properties=properties))
        label = self.get_namespaced_label(label=label, namespace=namespace)

        if properties is None:
            raise Exception("Vertex cannot be created with out  properties")
        _ = self.gremlin_client.g.addV(label)
        for k, v in properties.items():
            _.property(k, v)
        _vtx = _.elementMap().next()
        return VertexElement(_vtx, serializer=self.serializer)

    def get_or_create(self, label=None, namespace=None, properties=None):
        """

        :param label:
        :param namespace:
        :param properties:
        :return:
        """
        vertices = self.read_many(label=label, namespace=namespace, query=properties)
        print("vertices", vertices)
        if vertices.__len__() > 0:
            # return VertexElement(vertices[0], serializer=self.serializer)
            return vertices[0]
        else:
            return self.create(label=label, namespace=namespace, properties=properties)

    def update(self, vertex_id, properties=None):
        """

        :param vertex_id:
        :param properties:
        :return:
        """
        logger.debug("Updating vertex {vertex_id} with properties {properties}".format(
            vertex_id=vertex_id, properties=properties))
        properties = {} if properties is None else properties
        vtx_instance = self.filter_vertex(vertex_id=vertex_id)
        if vtx_instance is not None:
            for k, v in properties.items():
                vtx_instance.property(k, v)
            _vtx = vtx_instance.elementMap().next()
            return VertexElement(_vtx, serializer=self.serializer)
        return None

    def _read_one(self, vertex_id):
        """

        :param vertex_id:
        :return:
        """
        logger.debug("Finding vertex with id {vertex_id} ".format(vertex_id=vertex_id))
        filtered_data = self.gremlin_client.g.V(vertex_id)
        try:
            _ = filtered_data.elementMap().next()
            if _:
                return VertexElement(_, serializer=self.serializer)
        except Exception as e:
            print(e)
            pass
        return None

    def _read_many(self, label=None, namespace=None, query=None, limit=10, skip=0):
        logger.debug("Reading vertices with label:{label}; namespace:{namespace}; query: {query}; limit: {limit}"
                     "; skip: {skip}".format(label=label, namespace=namespace, query=query, limit=limit, skip=skip))
        filtered_data = self.filter_vertex(label=label, namespace=namespace, query=query, limit=limit, skip=skip)
        cleaned_data = []
        for _ in filtered_data.elementMap().toList():
            cleaned_data.append(
                VertexElement(_, serializer=self.serializer)
            )
        return cleaned_data

    def _delete_one(self, vertex_id):
        logger.debug("Deleting the vertex with vertex_id:{vertex_id}".format(vertex_id=vertex_id))
        self.drop(self.filter_vertex(vertex_id=vertex_id))

    def _delete_many(self, label=None, namespace=None, query=None):
        logger.debug("Deleting the vertex with label:{label}, namespace:{namespace}"
                     " query:{query}".format(label=label, namespace=namespace, query=query))
        self.drop(self.filter_vertex(label=label, namespace=namespace, query=query))

    def read_in_edges_and_vertices(self, vertex_id, label=None, namespace=None, query=None, limit=None, skip=None):
        # TODO - fix the performance, filter queries are made twice
        vtx = self.filter_vertex(vertex_id=vertex_id, label=label, namespace=namespace,
                                 query=query, limit=limit, skip=skip)
        cleaned_data = []
        for _ in vtx.inE().dedup().elementMap().toList():
            cleaned_data.append(EdgeElement(_, serializer=self.serializer))
        vtx = self.filter_vertex(vertex_id=vertex_id, label=label, namespace=namespace,
                                 query=query, limit=limit, skip=skip)
        for _ in vtx.inE().otherV().dedup().elementMap().toList():
            cleaned_data.append(VertexElement(_, serializer=self.serializer))
        return cleaned_data

    def read_out_edges_and_vertices(self, vertex_id, label=None, namespace=None, query=None, limit=None, skip=None):
        # TODO - fix the performance, filter queries are made twice
        vtx = self.filter_vertex(vertex_id=vertex_id, label=label, namespace=namespace,
                                 query=query, limit=limit, skip=skip)
        cleaned_data = []
        for _ in vtx.outE().dedup().elementMap().toList():
            cleaned_data.append(EdgeElement(_, serializer=self.serializer))
        vtx = self.filter_vertex(vertex_id=vertex_id, label=label, namespace=namespace,
                                 query=query, limit=limit, skip=skip)
        for _ in vtx.outE().otherV().dedup().elementMap().toList():
            cleaned_data.append(VertexElement(_, serializer=self.serializer))
        return cleaned_data

    def filter_vertex_and_neighbor_edges_and_vertices(self, vertex_id=None, label=None, namespace=None, query=None, limit=None,
                                        skip=None):
        cleaned_data = []
        vertices = self.filter_vertex(vertex_id=vertex_id, label=label, namespace=namespace,
                                      query=query, limit=limit, skip=skip)
        for _ in vertices.dedup().elementMap().toList():
            cleaned_data.append(VertexElement(_, serializer=self.serializer))

        # TODO - fix the performance, filter queries are made twice (damn! now increased to 5 times)
        in_edges_and_vertices = self.read_in_edges_and_vertices(vertex_id=vertex_id, label=label, namespace=namespace,
                                                                query=query, limit=limit, skip=skip)
        cleaned_data.extend(in_edges_and_vertices)
        out_edges_and_vertices = self.read_out_edges_and_vertices(vertex_id=vertex_id, label=label, namespace=namespace,
                                                                  query=query, limit=limit, skip=skip)
        cleaned_data.extend(out_edges_and_vertices)
        unique_elements = get_unique_items(cleaned_data)
        return unique_elements



from .base import GremlinOperationBase, CRUDOperationsBase
from gremlin_python.process.strategies import *
from gremlin_python.process.traversal import Order


class SchemaOps(GremlinOperationBase):

    def get_all_vertices_schema(self):
        _ = self.gremlin_client.execute_query(
            "g.V().group().by(label).by(properties().label().dedup().fold())",
            serialize_elements=False
        )
        schema_data = []
        for schema in _:
            for k, v in schema.items():
                schema_data.append(
                    {
                        "label": k,
                        "propertyKeys": v
                    }
                )
        return schema_data

    def get_vertex_label_schema(self, label: str, namespace: str = None):
        # TODO - fix performance
        _ = self.gremlin_client.execute_query(
            "g.V().group().by(label).by(properties().label().dedup().fold())",
            serialize_elements=False
        )
        return {"label": label, "propertyKeys": _[0].get(label, [])}

    def get_all_edges_schema(self):
        _ = self.gremlin_client.execute_query(
            "g.E().group().by(label).by(properties().label().dedup().fold())",
            serialize_elements=False
        )
        schema_data = []
        for schema in _:
            for k, v in schema.items():
                schema_data.append(
                    {
                        "label": k,
                        "propertyKeys": v
                    }
                )
        return schema_data

    def get_edge_label_schema(self, label: str, namespace: str = None):
        # TODO - fix performance
        _ = self.gremlin_client.execute_query(
            "g.E().group().by(label).by(properties().label().dedup().fold())",
            serialize_elements=False
        )
        return {"label": label, "propertyKeys": _[0].get(label, [])}

    def create_vertex_label_schema(self, label: str, namespace: str = None):

        try:
            _ = self.gremlin_client.execute_query(
                f"""
    mgmt = graph.openManagement()
    person = mgmt.makeVertexLabel('{label}').make()
    mgmt.commit()
                
                """,
                # "person = graph.addVertex(label, '" + label + "')",
                serialize_elements=False
            )
            return {"status": True, "message": "ok"}
        except Exception as e:
            return {"status": False, "message": e.__str__()}

    def create_edge_label_schema(self, label: str, multiplicity: str = None, namespace: str = None):
        # https://docs.janusgraph.org/basics/schema/#edge-label-multiplicity

        query = f"""
mgmt = graph.openManagement()
person = mgmt.makeEdgeLabel('{label}')"""
        if multiplicity:
            query += f".cardinality(Cardinality.{multiplicity.upper()})"
        query += ".make()"
        query += f"""
mgmt.commit()
        """
        try:
            _ = self.gremlin_client.execute_query(
                query,
                serialize_elements=False
            )
            return {"status": True, "message": "ok"}
        except Exception as e:
            return {"status": False, "message": e.__str__()}

    def create_vertex_property_schema(self,
                                      label: str,
                                      property_key: str,
                                      data_type: str,
                                      cardinality: str):
        """

        :param label:
        :param property_key:
        :param data_type:
        :param cardinality: SINGLE, LIST , SET
        :return:
        """

        query = f"""
        mgmt = graph.openManagement()
        {property_key}_prop = mgmt.makePropertyKey('{property_key}')
        """
        if data_type:
            query += f".dataType({data_type}.class)"
        if cardinality:
            query += f".cardinality(Cardinality.{cardinality.upper()})"
        query += ".make()"
        query += f"""
        
        {label}_label = mgmt.getVertexLabel("{label}")        
        mgmt.addProperties({label}_label, {property_key}_prop)
        mgmt.commit()
        """

        print("====", query)
        try:
            _ = self.gremlin_client.execute_query(
                query,
                serialize_elements=False
            )
            print("====", _)
            return {"status": True, "message": "ok"}
        except Exception as e:
            return {"status": False, "message": e.__str__()}

    def create_edge_property_schema(self,
                                    label: str,
                                    property_key: str,
                                    data_type: str,
                                    cardinality: str):
        """

        :param label:
        :param property_key:
        :param data_type:
        :param cardinality: SINGLE, LIST , SET
        :return:
        """

        query = f"""
        mgmt = graph.openManagement()
        {property_key}_prop = mgmt.makePropertyKey('{property_key}')
        """
        if data_type:
            query += f".dataType({data_type}.class)"
        if cardinality:
            query += f".cardinality(Cardinality.{cardinality.upper()})"
        query += ".make()"
        query += f"""
        
        {label}_label = mgmt.getEdgeLabel("{label}")        
        mgmt.addProperties({label}_label, {property_key}_prop)
        mgmt.commit()
        """
        try:
            _ = self.gremlin_client.execute_query(
                query,
                serialize_elements=False
            )
            print("====", _)
            return {"status": True, "message": "ok"}
        except Exception as e:
            return {"status": False, "message": e.__str__()}



from .base import CRUDOperationsBase
from .core.element import EdgeElement, VertexElement
import logging

logger = logging.getLogger(__name__)


class Edge(CRUDOperationsBase):

    def get_or_create(self, label=None, namespace=None, properties=None):
        """

        :param label:
        :param namespace:
        :param properties: {"id": 123213 }
        :return:
        """

        edges = self.read_many(label=label, namespace=namespace, query=properties)
        if edges.__len__() > 0:
            return edges[0]
        else:
            return self.create(label=label, namespace=namespace, properties=properties)

    def create(self, label=None, namespace=None, properties=None, inv=None, outv=None):
        """

        :param label:
        :param namespace:
        :param properties:
        :param inv: str or VertexElement
        :param outv: str or VertexElement
        :return:
        """
        logger.debug("Creating Edge with label {label}, namespace {namespace} and properties {properties}".format(
            label=label,
            namespace=namespace,
            properties=properties)
        )
        label = self.get_namespaced_label(label=label, namespace=namespace)
        properties = {} if properties is None else properties
        inv_vtx_instance = self.filter_vertex(inv)
        outv_vtx_instance = self.filter_vertex(outv)
        _ = inv_vtx_instance.addE(label) \
            .to(outv_vtx_instance)
        for property_key, property_value in properties.items():
            _.property(property_key, property_value)
        edg = _.elementMap().next()
        return EdgeElement(edg, serializer=self.serializer)

    def update(self, edg_id, properties=None):
        logger.debug("Updating vertex  {edg_id} with properties {properties}".format(edg_id=edg_id,
                                                                                     properties=properties, ))
        edge = self.filter_edge(edge_id=edg_id)
        properties = {} if properties is None else properties
        if edge:
            for k, v in properties.items():
                edge.property(k, v)
            _edge = edge.elementMap().next()
            return EdgeElement(_edge, serializer=self.serializer)
        return None

    def _read_one(self, edge_id):
        logger.debug("Finding edge with id {edge_id}".format(
            edge_id=edge_id))
        filtered_data = self.filter_edge(edge_id=edge_id)
        try:
            _ = filtered_data.elementMap().next()
            if _:
                return EdgeElement(_, serializer=self.serializer)
        except Exception as e:
            pass
        return None

    def _read_many(self, label=None, namespace=None, query=None, limit=10, skip=0):
        filtered_data = self.filter_edge(label=label, namespace=namespace, query=query, limit=limit, skip=skip)
        cleaned_data = []
        for _ in filtered_data.elementMap().toList():
            cleaned_data.append(EdgeElement(_, serializer=self.serializer))
        return cleaned_data

    def filter_edge_and_get_neighbor_vertices(self, edge_id=None, label=None, namespace=None, query=None, limit=None,
                                        skip=None):

        cleaned_edges_data = self._read_many(label=label, namespace=namespace,query=query, limit=limit, skip=skip)
        filtered_edges = self.filter_edge(label=label, namespace=namespace, query=query, limit=limit, skip=skip)

        vertices_data = []
        for _ in filtered_edges.inV().dedup().elementMap().toList():
            vertices_data.append(VertexElement(_, serializer=self.serializer))

        filtered_edges = self.filter_edge(label=label, namespace=namespace, query=query, limit=limit, skip=skip)

        for _ in filtered_edges.outV().dedup().elementMap().toList():
            vertices_data.append(VertexElement(_, serializer=self.serializer))
        vertices_data = list(set(vertices_data))

        return cleaned_edges_data + vertices_data


    def _delete_one(self, edge_id):
        logger.debug("Deleting the edge with edge_id:{edge_id}".format(edge_id=edge_id))
        self.drop(self.filter_edge(edge_id=edge_id))

    def _delete_many(self, label=None, namespace=None, query=None):
        logger.debug("Deleting the edges with label:{label} namespace:{namespace},"
                     " query:{query}".format(label=label, query=query, namespace=namespace))
        self.drop(self.filter_edge(label=label, namespace=namespace, query=query))



from .base import GremlinOperationBase, CRUDOperationsBase
from gremlin_python.process.strategies import *
from gremlin_python.process.traversal import Order
from invana_engine.server.settings import MANAGEMENT_VERTEX_NAME


class StatsOps(GremlinOperationBase):

    def get_vertices_label_stats(self, label: str = None, namespace: str = None, limit: int = None, order: str = None):
        stats = self.gremlin_client.g.V().label().groupCount().next()
        label_stats = []
        for label, count in stats.items():
            if label != MANAGEMENT_VERTEX_NAME:

                if namespace:
                    if "{}/".format(namespace) in label:
                        label_stats.append({"label": label, "count": count})
                else:
                    label_stats.append({"label": label, "count": count})

        return label_stats

    def get_edges_label_stats(self, label: str = None, namespace: str = None, limit: int = None, order: str = None):
        stats = self.gremlin_client.g.E().label().groupCount().next()
        label_stats = []
        for label, count in stats.items():
            if namespace:
                if "{}/".format(namespace) in label:
                    label_stats.append({"label": label, "count": count})
            else:
                label_stats.append({"label": label, "count": count})
        return label_stats

    def get_vertex_label_stats(self, label: str, namespace: str = None):
        _ = self.gremlin_client.g.V().hasLabel(label).count().next()
        return {"label": label, "count": _}

    def get_edge_label_stats(self, label: str, namespace: str = None):
        _ = self.gremlin_client.g.E().hasLabel(label).count().next()
        return {"label": label, "count": _}

    # def get_all_vertices_schema(self):
    #     _ = self.gremlin_client.execute_query(
    #         "g.V().group().by(label).by(properties().label().dedup().fold())",
    #         serialize_elements=False
    #     )
    #     schema_data = []
    #     for schema in _:
    #         for k, v in schema.items():
    #             schema_data.append(
    #                 {
    #                     "label": k,
    #                     "propertyKeys": v
    #                 }
    #             )
    #     return schema_data
    #
    # def get_vertex_label_schema(self, label: str, namespace: str = None):
    #     # TODO - fix performance
    #     _ = self.gremlin_client.execute_query(
    #         "g.V().group().by(label).by(properties().label().dedup().fold())",
    #         serialize_elements=False
    #     )
    #     return {"label": label, "propertyKeys": _[0].get(label, [])}
    #
    # def get_all_edges_schema(self):
    #     _ = self.gremlin_client.execute_query(
    #         "g.E().group().by(label).by(properties().label().dedup().fold())",
    #         serialize_elements=False
    #     )
    #     schema_data = []
    #     for schema in _:
    #         for k, v in schema.items():
    #             schema_data.append(
    #                 {
    #                     "label": k,
    #                     "propertyKeys": v
    #                 }
    #             )
    #     return schema_data
    #
    # def get_edge_label_schema(self, label: str, namespace: str = None):
    #     # TODO - fix performance
    #     _ = self.gremlin_client.execute_query(
    #         "g.E().group().by(label).by(properties().label().dedup().fold())",
    #         serialize_elements=False
    #     )
    #
    #     return {"label": label, "propertyKeys": _[0].get(label, [])}
    #



from urllib.parse import urlparse
import socket


def get_host(url):
    o = urlparse(url)
    return f"{o.scheme}://{o.netloc}"


def get_client_info():
    hostname = socket.gethostname()
    return {
        'host_name': hostname,
        'ip_address': socket.gethostbyname(hostname)
    }


def get_unique_items(elements):
    vertices = []
    edges = []
    existing_vertices = []
    existing_edges = []

    for elem in elements:
        elem_type = elem.type
        if elem_type == "g:Edge" and elem.id not in existing_edges:
            existing_edges.append(elem.id)
            edges.append(elem)
        elif elem_type == "g:Vertex" and elem.id not in existing_vertices:
            existing_vertices.append(elem.id)
            vertices.append(elem)
    return vertices + edges





import uvicorn
from invana_engine.server.app import app
import logging
from invana_engine.settings import server_port, shall_debug

if shall_debug is False:
    logging.getLogger('asyncio').setLevel(logging.ERROR)
    logging.getLogger('uvicorn').setLevel(logging.ERROR)
else:
    logging.basicConfig(level="DEBUG")


def server_start():
    uvicorn.run(app, host="0.0.0.0", port=int(server_port))


if __name__ == "__main__":
    server_start()



# from gremlin_python.driver.tornado.transport import TornadoTransport
# import asyncio
# import websockets
#
#
# class CustomTornadoTransport(TornadoTransport):
#
#     def __init__(self):
#         self._loop = asyncio.new_event_loop()
#
#     def connect(self, url, headers=None):
#         self._ws = self._loop.run_until_complete(
#             websockets.connect(url)
#         )
#
#     def write(self, message):
#         self._loop.run_until_complete(self._ws.send(message))
#
#     def read(self):
#         _ = self._loop.run_until_complete(self._ws.recv())
#         return _
#
#     def close(self):
#         self._ws.close()
#         self._loop.close()
#
#     def closed(self):
#         return not self._ws.protocol



from invana_engine.server.schemas.query import Gremlin
from graphene import Schema
import json
import os

schema = Schema(query=Gremlin)


class TestGraphQuery:

    @staticmethod
    def test_create_vertex():
        # label = "Planet"
        # properties = json.dumps({
        #     "name": "Earth",
        #     "human_population": "a lot"
        # })
        results = schema.execute("""
            query {
                vertexCreate(properties: "{\\"name\\": \\"Earth\\"}", label: "Planet")
            }
        """)
        print("results", results)
        # assert results.data == {"vertexCreate": "{\"name\": \"Earth\"}"}



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='invana-engine',
    version='0.0.1',
    description='GraphQL API and Insights engine for Apache TinkerPop supported graph databases.',
    author='Ravi Raja Merugu',
    author_email='ravi@invana.io',
    url='https://github.com/invanalabs/invana-engine',
    packages=find_packages(
        exclude=("dist", "docs", "tests", "scripts", "experiments")
    ),
    install_requires=[
        'gremlinpython==3.4.6',
        'starlette==0.13.8',
        'graphene==2.1.8',
        'uvicorn==0.12.2',
        'jinja2==2.11.2',
        'aiofiles==0.6.0'
    ],
    entry_points={
        'console_scripts': [
            'invana-engine-start = invana_engine.server_start:server_start',
        ]
    },
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Route
from starlette.graphql import GraphQLApp
from invana_engine.server.schemas.query import GremlinQuery
from .schemas.mutation import GremlinMutation
from graphene import Schema
from .views import homepage_view
from ..gremlin import GremlinClient
import time
from ..settings import gremlin_server_url, gremlin_server_password, gremlin_server_username, shall_debug

print(".................................................")
print("Starting Invana Engine server")
print(f"Using GREMLIN_SERVER_URL {gremlin_server_url}")
print(".................................................")

if gremlin_server_url is None:
    print("ERROR: GREMLIN_SERVER_URL environment variable not set. Please fix it .")
    print("Exiting the program now. Please refer the documentation at https://github.com/invanalabs/invana-engine")
    exit()

routes = [
    Route('/', endpoint=homepage_view),
    Route('/graphql', GraphQLApp(
        schema=Schema(query=GremlinQuery, mutation=GremlinMutation),
    ))
]

middleware = [
    Middleware(CORSMiddleware, allow_origins=['*'], allow_methods=["GET", "POST", "PUT", "DELETE"])
]

app = Starlette(routes=routes, middleware=middleware, debug=shall_debug)
time.sleep(1)
gremlin_client = GremlinClient(gremlin_server_url=gremlin_server_url,
                               gremlin_server_username=gremlin_server_username,
                               gremlin_server_password=gremlin_server_password
                               )
app.state.gremlin_client = gremlin_client



from .edge import Edge
from .vertex import Vertex
from .stats import StatsOps
from .schema import SchemaOps
from .serializers.graphson_v3 import GraphSONV3Reader
from gremlin_python.process.anonymous_traversal import traversal
from gremlin_python.driver.driver_remote_connection import DriverRemoteConnection
from gremlin_python.driver.protocol import GremlinServerWSProtocol, GremlinServerError
from gremlin_python.driver import request
from gremlin_python.driver.resultset import ResultSet

import base64


class InvanaGremlinServerWSProtocol(GremlinServerWSProtocol):

    def data_received(self, message, results_dict):
        # if Gremlin Server cuts off then we get a None for the message
        if message is None:
            raise GremlinServerError({'code': 500,
                                      'message': 'Server disconnected - please try to reconnect', 'attributes': {}})

        message = self._message_serializer.deserialize_message(message)
        request_id = message['requestId']
        result_set = results_dict[request_id] if request_id in results_dict else ResultSet(None, None)
        status_code = message['status']['code']
        aggregate_to = message['result']['meta'].get('aggregateTo', 'list')
        data = message['result']['data']
        result_set.aggregate_to = aggregate_to
        if status_code == 407:
            auth = b''.join([b'\x00', self._username.encode('utf-8'),
                             b'\x00', self._password.encode('utf-8')])
            request_message = request.RequestMessage(
                'traversal', 'authentication',
                {'sasl': base64.b64encode(auth).decode()})
            self.write(request_id, request_message)
            data = self._transport.read()
            # Allow recursive call for auth
            return self.data_received(data, results_dict)
        elif status_code == 204:
            result_set.stream.put_nowait([])
            del results_dict[request_id]
            return status_code
        elif status_code in [200, 206]:
            result_set.stream.put_nowait(data)
            if status_code == 200:
                result_set.status_attributes = message['status']['attributes']
                del results_dict[request_id]
            return status_code
        else:
            del results_dict[request_id]
            raise GremlinServerError(message["status"])


class _GremlinClient:
    """

    client = _GremlinClient("ws://127.0.0.1:8182/gremlin")
    client.execute_query("g.V().limit(1).toList()")
    """

    def __init__(self, gremlin_server_url, gremlin_server_username=None,
                 gremlin_server_password=None, serializer=None, transport_factory=None):
        if gremlin_server_url is None:
            raise Exception("Invalid gremlin_server_url. default: ws://127.0.0.1:8182/gremlin")
        self.connection = DriverRemoteConnection(
            gremlin_server_url, 'g',
            username=gremlin_server_username,
            password=gremlin_server_password,
            transport_factory=transport_factory
        )
        self.g = traversal().withRemote(self.connection)
        self.serializer = serializer

    def execute_query(self, raw_query, serialize_elements=True):
        """

        :param raw_query: Gremlin query in plain string.
        :param serialize_elements: this will convert data GraphSON data into JSON
        :return:
        """
        result = self.connection._client.submit(raw_query).all().result()
        if serialize_elements is True:
            return self.serializer.serialize_data(result)
        return result


class GremlinClient:
    """
    Usage:

    graph_client = GremlinClient(gremlin_server_url="ws://127.0.0.1:8182/gremlin")
    """

    def __init__(self, gremlin_server_url, gremlin_server_username=None,
                 gremlin_server_password=None, serializer=None, transport_factory=None):
        serializer = serializer or GraphSONV3Reader()
        self.gremlin_server_url = gremlin_server_url
        self.gremlin_client = _GremlinClient(gremlin_server_url=gremlin_server_url,
                                             gremlin_server_username=gremlin_server_username,
                                             gremlin_server_password=gremlin_server_password,
                                             serializer=serializer,
                                             transport_factory=transport_factory)
        self.vertex = Vertex(gremlin_client=self.gremlin_client)
        self.edge = Edge(gremlin_client=self.gremlin_client)
        self.stats = StatsOps(gremlin_client=self.gremlin_client)
        self.schema = SchemaOps(gremlin_client=self.gremlin_client)

    def close(self):
        self.gremlin_client.connection.close()

    def delete_everything(self):
        """

        :return: None
        """
        return self.gremlin_client.g.V().drop().iterate()

    def execute_query(self, raw_query,  serialize_elements=True):
        """

        :param raw_query: Gremlin query in plain string.
        :param serialize_elements: this will convert data GraphSON data into JSON
        :return:
        """
        return self.gremlin_client.execute_query(raw_query,  serialize_elements=serialize_elements)
